#include <hipc/atomic.h>
#include <hipc/queue.h>

/**
 * Initializes a queue.
 *
 * @param que_p pointer to a queue
 * @param buf_p a buffer whose size must be \c item_size multiplied by \c n_slot
 * @param item_size size of an item
 * @param n_slot number of slots
 */
void hipcQueueInit(struct HIPC_queue *const que_p,
                   void *const buf_p,
                   const size_t item_size, const size_t n_slot)
{
    que_p->buf_p = buf_p;
    que_p->item_size = item_size;
    que_p->n_slot = n_slot;
    que_p->head = 0;
    que_p->tail = 0;
    HIPC_MEM_BARRIER;

    return;
}

/**
 * Gets item size.
 *
 * @param que_p pointer to a queue
 * @returns item size
 */
size_t hipcQueueItemSize(struct HIPC_queue * const que_p)
{
    return que_p->item_size;
}

/**
 * Returns an address of the head item if it exists.
 *
 * @param que_p pointer to a queue
 * @returns \c NULL when the queue is empty,
 *          otherwise an address of an item to be popped is returned.
 */
void *hipcQueueEmpty(struct HIPC_queue *const que_p)
{
    if (que_p->head == que_p->tail) {
        return NULL;
    }

    return ((unsigned char *) (que_p->buf_p)) +
        (que_p->head * que_p->item_size);
}

/**
 * Pops an item and releases a slot.
 * Before each pop, it must be checked that a queue is not empty
 * by hipcQueueEmpty().
 *
 * @param que_p pointer to a queue
 */
void hipcQueuePop(struct HIPC_queue *const que_p)
{
    HIPC_MEM_BARRIER;
    que_p->head = (que_p->head + 1) % que_p->n_slot;

    return;
}

/**
 * Returns an address of an empty slot if it exists.
 *
 * @param que_p pointer to a queue
 * @returns \c NULL when the queue is full,
 *          otherwise an address of an item to be pushed is returned.
 */
void *hipcQueueFull(struct HIPC_queue const *const que_p)
{
    if (que_p->head == ((que_p->tail + 1) % que_p->n_slot)) {
        return NULL;
    }

    return ((unsigned char *) que_p->buf_p) +
        (que_p->tail * que_p->item_size);
}

/**
 * Pushes an item and occupies a slot.
 * Before each push, it must be checked that a queue is not full
 * by hipcQueueFull().
 *
 * @param que_p pointer to a queue
 */
void hipcQueuePush(struct HIPC_queue *const que_p)
{
    HIPC_MEM_BARRIER;
    que_p->tail = (que_p->tail + 1) % que_p->n_slot;

    return;
}

/**
 * Clears a queue.
 * Only a consumer, not a producer, should call this function because
 * the consumer may still use a slot.
 *
 * @param que_p pointer to a queue
 */
void hipcQueueClear(struct HIPC_queue *const que_p)
{
    HIPC_MEM_BARRIER;
    que_p->head = que_p->tail;

    return;
}
